#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <math.h>

#include <Adafruit_SSD1306.h>

#include "notes.h"
#include "song1.h"

#define SCREEN_WIDTH        128 // OLED display width, in pixels
#define SCREEN_HEIGHT        64 // OLED display height, in pixels

#define SPEAKER_PIN 9

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET            4 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

const byte buttonPins[] = {4, 2, 5, 3}; // LEFT, UP, RIGHT, DOWN

const int line_positions[] = {23,53, 83, 113};

bool button1_down = false;
bool button2_down = false;
bool button3_down = false;
bool button4_down = false;


typedef enum {
  START,
  RUNNING,
  GAMEOVER
} State;

typedef struct Note {
    int midi_note;           // The integer value
    double delta; // The delta time in milliseconds or microseconds
} Note;


typedef struct FallingNote {
    int string_num;      
    double hit_time; // The delta time in when note actually occurs
    double currentY; // where note currently is 
} FallingNote;

#define MAX_TUPLES 150 // Define the maximum number of tuples
Note notes[MAX_TUPLES];
Note* copyNotes;

double oldYs[MAX_TUPLES];

Note string1[MAX_TUPLES];
int s1_notes = 0;
Note string2[MAX_TUPLES];
int s2_notes = 0;
Note string3[MAX_TUPLES];
int s3_notes = 0;
Note string4[MAX_TUPLES];
int s4_notes = 0;

FallingNote falling_notes[MAX_TUPLES];

// for notes that fall
int finalY = SCREEN_HEIGHT-7;
int fallingSpeed = 20; // in pixels per second  


int noteCount = 0; // Keep track of the number of stored tuples
int currentNote = 0; 

const float BPM = 141;

State gameState;

int score = 0;
int highScore = 0;

// when notes begin
const int TIME_OFFSET = 3;
unsigned long startingTime;
unsigned long currentMilis;

void setup() {
  Serial.begin(9600);

  if(!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;);
  }

  for (byte i = 0; i < 4; i++) {
    pinMode(buttonPins[i], INPUT_PULLUP);
  }

  randomSeed(analogRead(A0));

  pinMode(SPEAKER_PIN, OUTPUT);

  setupGame();

  // prepare notes for song
  populate_note_array();

  // copy song for putting on strings 
  copyNotes = (Note*)malloc(noteCount * sizeof(Note));
  memcpy(copyNotes, notes, noteCount * sizeof(Note));

  qsort(copyNotes, noteCount, sizeof(Note), compare);

  bucket(copyNotes);


}

void setupGame() {
  gameState = START;

  display.clearDisplay();
  drawPressToStart();
  display.display();
}


int moveTime = 0;
void loop() {
 switch(gameState) {
    case START:
      if(buttonPress()) 
      { 
        gameState = RUNNING; 
        setupLines();   
        calculateStartingY();
       // printYPos();
        startingTime = millis(); 
        drawDots();
      }

      break;
      
    case GAMEOVER:
      display.clearDisplay();
      drawGameover();
      if(buttonPress()) {

        highScore = score;
        score = 0;


        gameState = RUNNING;
        setupLines();   
        calculateStartingY();
        printYPos();
        startingTime = millis(); 
        drawDots();
        currentNote = 0;

      }
      break;

    case RUNNING:
      unsigned long currentMillis = millis();
      if(currentMillis-startingTime >= notes[currentNote].delta*1000 && currentNote < noteCount)
      {
        // Serial.println(currentNote);
        // Serial.println(noteCount);

        playNextNote();
      }
      readSpecificButtonPress();

      calculateNewPos((currentMillis-startingTime)/1000.0);
      drawDots();

      if(currentNote >= noteCount)
      {
        delay(1000);
        gameState = GAMEOVER;

      }

        
      drawScore();

      break;
  }

  delay(10);
}

bool buttonPress() {
  for (byte i = 0; i < 4; i++) {
    byte buttonPin = buttonPins[i];
    if (digitalRead(buttonPin) == LOW) {
      return true;
    }
  }
  return false;
}

void readSpecificButtonPress() {
  
  for (byte i = 0; i < 4; i++) {
    byte buttonPin = buttonPins[i];
    if (digitalRead(buttonPin) == LOW) {
      display.fillRect(line_positions[i] - 5, SCREEN_HEIGHT - 10, 10, 10, WHITE);
      if(i == 0)
        button1_down = true;
      if(i == 1)
        button2_down = true;
      if(i == 2)
        button3_down = true;
      if(i == 3)
        button4_down = true;
    }
    else
    {
      if(i == 0)
        button1_down = false;
      if(i == 1)
        button2_down = false;
      if(i == 2)
        button3_down = false;
      if(i == 3)
        button4_down = false;
        // clear old rectangle
        display.fillRect(line_positions[i] - 5, SCREEN_HEIGHT - 10, 10, 10, BLACK);
        display.drawRect(line_positions[i] - 5, SCREEN_HEIGHT - 10, 10, 10, WHITE);
    }
  }
  display.display();
}



void setupLines()
{
  display.clearDisplay();
  drawScore();
  for(int i = 0; i < 4; i++)
  {
    display.drawLine(line_positions[i], 0, line_positions[i], SCREEN_HEIGHT-10, WHITE);
    display.drawRect(line_positions[i] - 5, SCREEN_HEIGHT - 10, 10, 10, WHITE);
  }
  
  display.display();
}

void drawScore() {


 display.fillRect(0, 0, 20, 30, SSD1306_BLACK); 

  display.setTextSize(0.5);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0,2);
  display.println(F("Scr:"));
  display.println(score);


}

void drawPressToStart() {
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(2,20);
  display.print(F("PRESS TO \nSTART"));
}

void drawGameover() {
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(1,5);
  display.println(F("GAMEOVER"));
  display.println(F("SCORE: "));
  display.println(score);
  if(score > highScore)
  {
     display.println(F("NEW HIGH SCORE!\nOld HIGH SCORE WAS: "));
      display.println(highScore);
  
  }

  display.println("PRESS TO PLAY AGAIN");

  display.display();
     
}

// audio code

void playNextNote()
{
  // all note past first need to also stop previous note
  // if(currentNote != 0)
  // { 
  //   noTone(SPEAKER_PIN);
  // }
  // stop note after time of note value finishes minus a bit for attack
  tone(SPEAKER_PIN, midiToFrequency(notes[currentNote].midi_note),song[currentNote][1]*(1.0/8.0)*(60000/BPM));
  // Serial.println(currentNote);
  currentNote++;
}

void populate_note_array()
{

    int startingTime = TIME_OFFSET;
    int currentTime = startingTime;

    double numberBeatsBefore = 0;

    // get first note 
    notes[0].midi_note = song[0][0];
    notes[0].delta = startingTime;
    noteCount++;

    for(int i = 1; i < songOneNotes; i++) {
        notes[noteCount].midi_note = song[i][0];
        numberBeatsBefore += song[noteCount-1][1]/4.0; 

        notes[noteCount].delta = numberBeatsBefore*(60.0/BPM)+TIME_OFFSET;

       // Serial.println(notes[noteCount].midi_note);
      //  Serial.println(notes[noteCount].delta);
        noteCount++;
       // Serial.println();
    }
    // Serial.println("FINISHED LOADING SONG");
}

double midiToFrequency(int midiNote) {
    return 440.0 * pow(2.0, (midiNote - 69) / 12.0);
}


int compare(const void* a, const void* b) {
    Note* structA = (Note*)a;
    Note* structB = (Note*)b;
  return structA->midi_note - structB->midi_note;
}


// divide song into strings for sending down each line
void bucket(Note*copyNotes)
{
  int bucketSize = noteCount/4;

  int counter =  0;

    for(int y  = 0; y < bucketSize; y++ )
    {
      string1[y].midi_note = copyNotes[counter].midi_note;
      string1[y].delta = copyNotes[counter].delta;
      s1_notes++;

      falling_notes[counter].string_num = 1;
      falling_notes[counter].hit_time = copyNotes[counter].delta;;
      // calculate this later 
      falling_notes[counter].currentY = -1;

      counter++;
    }
    for(int y  = 0; y < bucketSize; y++ )
    {
      string2[y].midi_note = copyNotes[counter].midi_note;
      string2[y].delta = copyNotes[counter].delta;
      s2_notes++;

       falling_notes[counter].string_num = 2;
      falling_notes[counter].hit_time = copyNotes[counter].delta;;
      // calculate this later 
      falling_notes[counter].currentY = -1;
      counter++;
    }
    for(int y  = 0; y < bucketSize; y++ )
    {
      string3[y].midi_note = copyNotes[counter].midi_note;
      string3[y].delta = copyNotes[counter].delta;
      s3_notes++;

       falling_notes[counter].string_num = 3;
      falling_notes[counter].hit_time = copyNotes[counter].delta;
      // calculate this later 
      falling_notes[counter].currentY = -1;
      counter++;
    }

    // final bucket smaller
    int notesLeft = noteCount-bucketSize*3;
    for(int y  = 0; y < notesLeft; y++ )
    {
      string4[y].midi_note = copyNotes[counter].midi_note;
      string4[y].delta = copyNotes[counter].delta;
      s4_notes++;

       falling_notes[counter].string_num = 4;
      falling_notes[counter].hit_time = copyNotes[counter].delta;;
      // calculate this later 
      falling_notes[counter].currentY = -1;
      counter++;
    }

}

void printNoteArray(int noteCount, Note*array)
{
  for(int i = 0; i < noteCount; i++)
  {
    Serial.println(array[i].midi_note);
  }
}

// calc where each circle begins given when it needs to hit 
void calculateStartingY()
{
  for(int i = 0; i < noteCount; i++)
  {
    oldYs[i] = falling_notes[i].currentY;
    falling_notes[i].currentY = calcNotePos(falling_notes[i].hit_time, 0);
  }
}

// calculate wher note should be on screen given any time 
double calcNotePos(double noteTimestamp, double currentTime)
{
  return -(fallingSpeed*(noteTimestamp-currentTime)-finalY);
}


void printYPos()
{
  for(int i = 0; i < noteCount; i++)
  {
  Serial.print(falling_notes[i].currentY);
  Serial.print(" ");
  Serial.println(falling_notes[i].hit_time);
  }
}

void drawDots()
{

  for(int i =0; i < noteCount; i++)
  {
    // black out old position
    double xCoor = line_positions[falling_notes[i].string_num-1];
    // redo old lines 
    for(int i = 0; i < 4; i++)
    {
    display.drawLine(line_positions[i], 0, line_positions[i], SCREEN_HEIGHT-10, WHITE);
    }

    display.fillCircle(xCoor, oldYs[i], 3,  BLACK);
    // draw new circle
    display.fillCircle(xCoor, falling_notes[i].currentY, 3,  WHITE);
    // check if circle is in square and button is held, add to score
    if(falling_notes[i].string_num == 1 && button1_down && (falling_notes[i].currentY <= SCREEN_HEIGHT && falling_notes[i].currentY >= SCREEN_HEIGHT-10))
      score += 5;
    if(falling_notes[i].string_num == 2 && button2_down && (falling_notes[i].currentY <= SCREEN_HEIGHT && falling_notes[i].currentY >= SCREEN_HEIGHT-10))
      score += 5;
    if(falling_notes[i].string_num == 3 && button3_down && (falling_notes[i].currentY <= SCREEN_HEIGHT && falling_notes[i].currentY >= SCREEN_HEIGHT-10))
      score += 5;
    if(falling_notes[i].string_num == 4 && button4_down && (falling_notes[i].currentY <= SCREEN_HEIGHT && falling_notes[i].currentY >= SCREEN_HEIGHT-10))
      score += 5;

  }
  
}

// calculate dots new pos at current time 
void calculateNewPos(double currTime)
{
 for(int i = 0; i < noteCount; i++)
  {
    oldYs[i] = falling_notes[i].currentY;
    falling_notes[i].currentY = calcNotePos(falling_notes[i].hit_time, currTime);
  }

}

